/**
 * 
 */
package de.ppi.psec.helper;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import play.Logger;
import play.Logger.ALogger;
import play.Play;
import play.i18n.Messages;
import play.mvc.Action;
import play.mvc.Http.Context;
import play.mvc.Http.Request;
import play.mvc.Result;
import play.mvc.Results;
import controllers.psec.AuthController;
import de.ppi.psec.permissions.IsPublic;
import de.ppi.psec.services.AuthenticationService;
import de.ppi.psec.services.AuthenticationServiceImpl;
import de.ppi.psec.services.AuthorizationServiceImpl;
import de.ppi.psec.services.PublicActionsByConfiguration;


/**
 * Helper class to protect access.
 * @author niels
 *
 */
public class PsecAction extends Action.Simple {
    
    private final static ALogger log = Logger.of(PsecAction.class);
         
    private static Map<Method, Boolean> publicMethod = new HashMap <Method, Boolean>();
    
    private final String loginUrl;
    
    private final boolean isEnabled;
    
    private final Method actionMethod;
    
    
    /**
     * Constructor    
     * @param isEnabled true if the protection should be enabled.
     * @param loginUrl the url for the login.
     */
    public PsecAction(boolean isEnabled, String loginUrl, Method actionMethod) {
        super();
        this.isEnabled = isEnabled;
        this.loginUrl = loginUrl;
        this.actionMethod = actionMethod;
    }


    @Override
    public Result call(Context ctx) throws Throwable {
        Context.current.set(ctx);
        @SuppressWarnings("rawtypes")
        final Class controllerClass = actionMethod.getDeclaringClass();
        final String action = actionMethod.getName();
        final Request request = ctx.request();
        log.trace("Check controller >" + controllerClass+ "< and action >" + 
                action + "<");
        if (!AuthController.class.equals(controllerClass)) {
            ctx.session().remove(AuthController.FLASH_PSEC_URL);
        }
        if (isEnabled && !isPublicMethod(actionMethod)) {
           AuthenticationService authService = AuthenticationServiceImpl.getService();
           if (!authService.isAuthenticate(true) && !authService.canRemembered()) {
               final String interruptedUrl = getInterruptedUrl(request);
               ctx.session().put(AuthController.FLASH_PSEC_URL, interruptedUrl);
               return Results.temporaryRedirect(loginUrl);
           }
           final String target = controllerClass.getName();
           if (!AuthorizationServiceImpl.getService().
                   hasPermission(target, action)) {
               if (Play.isProd()) {
                   return Results.forbidden(Messages.get("access.denied.prod", target, action));
               } else {
                   return Results.forbidden(Messages.get("access.denied.debug", target, action));
               }
           }
        }
        if (AuthController.class.equals(controllerClass) && 
                !ctx.session().containsKey(AuthController.FLASH_PSEC_URL)) {
            final String lastUrl = getLastURL(request);
            ctx.session().put(AuthController.FLASH_PSEC_URL, lastUrl);                
        }
        return delegate.call(ctx);
    }


    
    /**
     * Checks if the given method is public or not.
     * @param method the method
     * @return true - if the method is public.
     */
    public boolean isPublicMethod(Method method) {
        if (PublicActionsByConfiguration.isPublic(method)) {
            return true;
        }
        if (!publicMethod.containsKey(method)) {
            if (method.isAnnotationPresent(IsPublic.class) || method.
                    getDeclaringClass().isAnnotationPresent(IsPublic.class)) {
                publicMethod.put(method, Boolean.TRUE);
            } else {
                publicMethod.put(method, Boolean.FALSE);
            }
        }
        return publicMethod.get(method).booleanValue();
    }


    
    /**
     * Grep the URL which was interrupted to make a good redirect, after login.
     * @param request the Request.
     * @return the URL.
     */
    private String getInterruptedUrl(final Request request) {
        String interruptedUrl;
        if ("GET".equals(request.method())) {
            interruptedUrl = request.uri();
        } else {
            interruptedUrl = getLastURL(request);
        }
        if (log.isDebugEnabled()) log.debug(request.method() + ":" + interruptedUrl);
        return interruptedUrl;
    }


    private String getLastURL(final Request request) {
        String interruptedUrl;
        final String referer = request.getHeader("referer");
        if (referer != null && referer.length() > 1 
                && !referer.endsWith(request.uri())) {
            interruptedUrl = referer;
        } else {
            interruptedUrl = "/";
        }
        if (log.isDebugEnabled()) log.debug("LastURL:" + interruptedUrl);
        return interruptedUrl;
    }


}
