/*
 * Copyright 2009 PPI-AG
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.ppi.psec.permissions;



import java.security.Permission;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * A basic permission class that can be used with the default DB realm
 * provided with the plugin.
 */
public class PsecBasicPermission extends Permission {
    public static final String WILDCARD = "*";

    private String target;
    private Set<String> actions;
    private String actionsString;

    /**
     * Creates a new permission with the given actions. <b>Note</b> an
     * action may not have any whitespace in it, or ',' or ';' characters.
     * @param target The permission target. The meaning of 'target' is
     * application dependent, for example it may be a file path, a URL,
     * a controller name, or a component name.
     * @param actions A string of actions separated by ',', ';', or
     * whitespace. For example, 'view,edit', 'create, delete',
     * 'create;    modify', 'view edit' are all valid two-action strings.
     */
    public PsecBasicPermission(String target, String actions) {
        super(actions+"@"+target);
        this.target = target;

        // Convert the delimited string of actions into a set.
        this.actions = actionsStringToSet(actions);
    }

    /**
     * Creates a new permission with the given actions.
     * @param target The permission target. The meaning of 'target' is
     * application dependent, for example it may be a file path, a URL,
     * a controller name, or a component name.
     * @param actions A collection of action strings, for example 'view'
     * 'edit', 'create', and 'delete'.
     */
    public PsecBasicPermission(String target, Collection<String> actions) {
        super(actionsToString(actions)+"@"+target);
        this.target = target;

        // Copy the provided collection of actions.
        this.actions = new HashSet<String>(actions);
    }

    /**
     * Returns an unsmodifiable set of this permission's actions.
     */
    public Set<String> getActionsSet() {
        return Collections.unmodifiableSet(this.actions);
    }

    /**
     * Returns this permission's actions as a comma-separated string,
     * for example 'view,create,modify,delete'.
     */
    @Override
    public String getActions() {
        // Delayed initialisation of the actions string.
        if (this.actionsString == null) {
            // Convert the set of actions to a string so that a standard
            // delimiter is always used.
            this.actionsString = actionsToString(this.actions);
        }

        return this.actionsString;
    }

    
    /**
     * Returns the permission target. The meaning of 'target' is
     * application dependent, for example it may be a file path, a URL,
     * a controller name, or a component name.
     * @return the permission target.
     */
    public String getTarget() {
        return target;
    }

    /**
     * Returns <code>true</code> if this permission implies the given
     * one, otherwise <code>false</code>. This method checks that this
     * permission has the same target as the given permission and a
     * super-set of the actions. The wildcard matches all targets and
     * all actions.
     */
    @Override
    public boolean implies(Permission p) {
        if (this==p) {
            return true;
        }

        boolean implies = (p instanceof PsecBasicPermission);

        if (implies) {
            // Check the target strings are the same.
            PsecBasicPermission perm = (PsecBasicPermission) p;
            if (this.target == null) {
                implies = (perm.target == null);
            } else if (!this.target.equals(WILDCARD)) {
                implies = this.target.equals(perm.target);
            }

            // Now check the actions (unless the current permission
            // includes the wildcard, which means that is implies
            // all other actions).
            if (implies && !this.actions.contains(WILDCARD)) {
                implies = this.actions.containsAll(perm.actions);
            }
        }
        return implies;
    }

    /**
     * Perfomance-Function which makes the check for target + action
     * faster.
     * @param requiredTarget the target which needs permission
     * @param requiredAction the action which needs permission
     * @return true if this permission allow the given target and action.
     */
    public boolean implies(String requiredTarget, String requiredAction) {
        boolean implies = true;
        if (this.target == null) {
            implies = requiredTarget == null;
        } else if (!this.target.equals(WILDCARD)) {
            implies = this.target.equals(requiredTarget);
        }

        // Now check the actions (unless the current permission
        // includes the wildcard, which means that is implies
        // all other actions).
        if (implies && !this.actions.contains(WILDCARD)) {
            implies = this.actions.contains(requiredAction);
        }
        return implies;
    }

    // Custom equals() method.
    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof PsecBasicPermission)) {
            return false;
        }

        PsecBasicPermission p = (PsecBasicPermission) obj;
        return (this.target == null ? p.target == null : this.target.equals(p.target)) &&
                this.actions.equals(p.actions);
    }

    // Custom hashCode() method.
    @Override
    public int hashCode() {
        int result;
        result = (this.target != null ? this.target.hashCode() : 0);
        result = 31 * result + this.actions.hashCode();
        return result;
    }

    /**
     * Converts a delimiter-separated string of actions into a set of
     * actions. Supported delimiters are ',', ';', and whitespace.
     */
    protected Set<String> actionsStringToSet(String actions) {
        return new HashSet<String>(Arrays.asList(actions.split("[,;\\s][\\s]*")));
    }

    /**
     * Converts a collection of actions into a comma-separated string.
     * For example, 'view,edit,create,delete'. Note that there is no
     * whitespace in the resulting string.
     */
    static private String actionsToString(Collection<String> actions) {
        StringBuffer buffer = new StringBuffer(actions.size() * 10);
        for (Iterator<String> iter = actions.iterator(); iter.hasNext();) {
            buffer.append(iter.next()).append(',');
        }

        return buffer.substring(0, buffer.length() - 1);
    }

    @Override
    public String toString() {
        return "PsecBasicPermission [target=" + target + ", getActions()="
                + getActions() + "]";
    }


}
