package de.ppi.psec.services;

import java.io.UnsupportedEncodingException;
import java.security.Permission;
import java.util.ArrayList;
import java.util.List;


import models.psec.User;
import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;
import net.sf.ehcache.Status;
import play.Logger;
import play.Logger.ALogger;
import play.Play;
import play.libs.Time;
import play.mvc.Http.Context;
import play.mvc.Http.Cookie;
import play.mvc.Http.Cookies;
import play.mvc.Http.Request;
import play.mvc.Http.Response;
import play.mvc.Http.Session;
import de.ppi.util.Crypto;
import de.ppi.util.PasswordTools;



public class AuthenticationServiceImpl implements AuthenticationService {

    private static final String COOKIE_NAME = "psec_rememberMe";
    private static final String SESSION_KEY_USERID = "psecUid";
    private static final String SESSION_KEY_REMEMBERED = "psecRem";

    private static final ALogger log = Logger.of(AuthenticationServiceImpl.class);

    private static AuthenticationService service;
    
    private final UserPersistanceService userPersistanceService;

    private final int maxLoginFailed;

    private final String rememberMeCryptKey;
    
    /** 
     * This cache contains all information, which hangs directly at the user,
     * like name, roles and direct permission.
     */
    private Cache cache;

    /**
     * Number of seconds the values are stored in the cache.
     */
    private final int cacheExpiration = 60 * 60;

    private final int maxCookieAge;


    /**
     * Creates an {@link AuthenticationService}.
     * @param maxLoginFailed maximal numbers of failed logins.
     * @param maxCookieAgeString age of the rememberMe-Cookie in format of {@link Time#parseDuration(String)}
     * @param rememberMeCryptKey the key to encrypt rememberMe-data.
     * @param userPersistanceService a service to read and safe user-data.
     */
    public AuthenticationServiceImpl(int maxLoginFailed,
            String maxCookieAgeString, String rememberMeCryptKey, 
            UserPersistanceService userPersistanceService) {
        super();
        this.maxLoginFailed = maxLoginFailed;
        this.rememberMeCryptKey = rememberMeCryptKey.substring(0, 16);
        maxCookieAge = Time.parseDuration(maxCookieAgeString);
        getCache();
        this.userPersistanceService = userPersistanceService;
    }




    /**
     * {@inheritDoc}
     */
    public boolean isAuthenticate(boolean allowRememberMe) {
        final Session session = getSession();
        if (getUserid() != null) {
            if (allowRememberMe) {
                return true;
            } else {
                if (session.get(SESSION_KEY_REMEMBERED) == null){
                    return true;
                } else {
                    return !"Y".equals(session.get(SESSION_KEY_REMEMBERED));
                }
            }
        }
        return false;
    }





    /**
     * {@inheritDoc}
     */
    public boolean canRemembered() {
        final Session session = getSession();
        final Request request = getRequest();
        final Response response = getResponse();
        final Cookie cookie = getRememberMeCookieFromRequest();
        if (cookie != null && cookie.value() != null && cookie.value().length() > 3) {
            final String cookieValue = Crypto.decryptAES(cookie.value(), rememberMeCryptKey);
            int delimiterPos = cookieValue.indexOf('|');
            boolean javaScriptOn = (cookieValue.startsWith("Y"));
            Long userId = Long.valueOf(cookieValue.substring(1,
                    delimiterPos));
            String hashWert = cookieValue.substring(delimiterPos+1);
            boolean remembered;
            final User user = userPersistanceService.findUserById(userId);
            if (user == null) {
                log.warn("User with id " + userId + "not founded.");
                remembered = false;
            } else if (user.getLoginFailedCount() > maxLoginFailed || !user.isEnabled()) {
                log.debug("User " + user.getId() + " founded but not active.");
                remembered = false;
            } else {
                if (log.isInfoEnabled()) {
                    log.info(String.format("Remember user '%s'  uses %s JavaScript: %b",
                        user.getLoginname(), request.headers().get("User-Agent"), 
                        Boolean.valueOf(javaScriptOn)));
                }
                // Now check the user's password against the hashed value stored
                // in the database.
                boolean passwordValid = false;
                try {
                    passwordValid = PasswordTools.checkDigestBase64(createCookiePassword(user),hashWert);
                } catch (UnsupportedEncodingException e) {
                    log.warn("Couldn't check remember me password", e);
                }
                if (!passwordValid) {
                    log.info("Invalid cookiepassword for '" + user.getLoginname() + "'");
                    user.setLoginFailedCount(user.getLoginFailedCount() + 1);
                    if (user.getLoginFailedCount() > maxLoginFailed) {
                        log.warn("Invalid Cookie! '" + user.getLoginname() + "' is now disabled.");
                    }
                    userPersistanceService.saveUser(user);
                    response.discardCookies(COOKIE_NAME);
                    remembered = false;
                } else {
                    setSessionParameters(user, session);
                    remembered = true;
                }
            }
            if (remembered) {
                session.put(SESSION_KEY_REMEMBERED, "Y");
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }

    }




    /**
     * {@inheritDoc}
     */
    public boolean login(String loginname, String password, boolean rememberMe,
            boolean javaScriptOn) {
        final Session session = getSession();
        final Request request = getRequest();
        final Response response = getResponse();

        boolean result = true;
        if (loginname == null || password == null) {
            result = false;
        } else {
            final String ipInfo = "unknown";//TODO niels request.remoteAddress;
            final User user = userPersistanceService.findUserByLoginname(loginname);
            if (user == null) {
                log.warn("Login failed unknown user " + loginname +  " from " + ipInfo);
                result = false;

            } else if (user.getLoginFailedCount() > maxLoginFailed || !user.isEnabled()) {
                log.warn("Login failed for user " + user.getLoginname() +
                        " from " + ipInfo +", because the user is disabled.");
                result = false;
            } else {
                log.info("Found user '" + user.getLoginname() + "' in DB");
                // Now check the user's password against the hashed value stored
                // in the database.
                if (!user.checkPassword(password)) {
                    log.info("Invalid password for '" + user.getLoginname() +
                            "' from " + ipInfo);
                    user.setLoginFailedCount(user.getLoginFailedCount() + 1);
                    if (user.getLoginFailedCount() > maxLoginFailed) {
                        log.warn("Invalid Password! '" + user.getLoginname() +
                                "' is now disabled.From " + ipInfo);
                    }
                    userPersistanceService.saveUser(user);
                    result = false;
                } else {
                    if (user.getLoginFailedCount() > 0) {
                        user.setLoginFailedCount(0);
                        userPersistanceService.saveUser(user);
                    }
                    //If a user has a login kill his remember me and session-info.
                    logout();
                    setSessionParameters(user, session);
                    if (rememberMe) {
                        setRemembeMeCookie(user, javaScriptOn,
                        		//TODO wie kriege ich secure() raus. request.secure().booleanValue()
                        		true, response);
                    }
                    result = true;
                }
            }
        }
        log.info (loginname + " uses " + request.headers().get("User-Agent") +
                " JavaScript: " + javaScriptOn + " login:" + result);
        return result;
    }

    
    /**
     * {@inheritDoc}
     */
    public void logout() {
        final Session session = getSession();
        final Response response = getResponse();
        session.remove(SESSION_KEY_USERID);
        if (getRememberMeCookieFromRequest() != null) {
            response.discardCookies(COOKIE_NAME);
        }
    }

    private Cookie getRememberMeCookieFromRequest() {
        Cookies cookies = getRequest().cookies();  
        return cookies.get(COOKIE_NAME);
    }
    

    /**
     * {@inheritDoc}
     */
    public String getLoginname() {
        String keyName = ".loginname";
        return (String)getCacheValue(keyName);
    }

    /**
     * {@inheritDoc}
     */
    public String getUserid() {
        return getSession().get(SESSION_KEY_USERID);
    }

    /**
     * {@inheritDoc}
     */
    public String getLastname() {
        String keyName = ".lastname";
        return (String)getCacheValue(keyName);
    }

    /**
     * {@inheritDoc}
     */
    public String getFirstname() {
        String keyName = ".firstname";
        return (String)getCacheValue(keyName);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public List<String> getRoles() {
        String keyName = ".roles";
        return (List<String>)getCacheValue(keyName);
    }
    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public List<Permission> getPermissions() {
        String keyName = ".permissions";
        return (List<Permission>)getCacheValue(keyName);
    }

    @Override
    public void clearCache() {
        log.info("Clear authentication-cache");
        getCache().removeAll();
    }

    private void setRemembeMeCookie(User user, boolean javaScriptOn,
            boolean secure, Response response) {
        String javaScriptOnChar = javaScriptOn?"Y":"N";
        String rememberMeKey = javaScriptOnChar + user.getId() + "|" +
                PasswordTools.saltPasswordBase64(createCookiePassword(user));
        String cookieValue = "";
        cookieValue = Crypto.encryptAES(rememberMeKey, rememberMeCryptKey);
        response.setCookie(COOKIE_NAME, cookieValue, maxCookieAge, "/", null, secure, true);
    }


    private void setSessionParameters(User user, Session session) {
    	session.put(SESSION_KEY_USERID, user.getId().toString());
        defineCache(user);
    }

    private void defineCache(String userId) {
        User user = findUserById(userId);
        defineCache(user);
    }

    private User findUserById(String userId) {
        return userPersistanceService.findUserById(Long.valueOf(userId));
    }
    
    
    
    private void defineCache(User user) {
        setCacheValue(user.getId() + ".loginname", user.getLoginname(), cacheExpiration);
        setCacheValue(user.getId() + ".lastname", user.getLastname(), cacheExpiration);
        setCacheValue(user.getId() + ".firstname", user.getFirstname(), cacheExpiration);
        setCacheValue(user.getId() + ".roles", user.getRoles(), cacheExpiration);
        final List<Permission> permissions = new ArrayList<Permission>();
        if (user.getPermissions() != null)  {
            permissions.addAll(user.getPermissions());
        }
        setCacheValue(user.getId() + ".permissions", permissions, cacheExpiration);
    }

    /**
     * Returns the value from the cache. If the cache is empty it will
     * set the values to the cache again.
     * @param keyName name of the key to access the cache.
     * @return the value stored under the keyName.
     */
    private Object getCacheValue(String keyName) {
        final String userId = getUserid();
        Element keyvalue = getCache().get(userId + keyName);
        if (keyvalue == null) {
            defineCache(userId);
            keyvalue = getCache().get(userId + keyName);
        }
        return keyvalue.getValue();
    }


    private String createCookiePassword(User user) {
        return user.getLoginname() + ":" + user.getPasswordHash();
    }




    public static AuthenticationService getService() {
        if (service == null) {
            final Integer maxLoginConfigValue = Play.application().configuration().
                    getInt("permsec.maxLoginFailed");
            String maxCookieAgeConfig = Play.application().configuration().
                    getString("permsec.maxCookieAge");
            String rememberMeCryptKeyConfig = Play.application().configuration().
                    getString("permsec.rememberMeCryptKey");
            if (rememberMeCryptKeyConfig == null) {
                rememberMeCryptKeyConfig = Play.application().configuration().
                        getString("application.secret");
            }
            service = new AuthenticationServiceImpl(maxLoginConfigValue.intValue(), 
                    maxCookieAgeConfig, rememberMeCryptKeyConfig, new UserPersistanceServiceEbean());
        }
        return service;
    }

    public synchronized static void setService(AuthenticationService service) {
        AuthenticationServiceImpl.service = service;
    }

    private void setCacheValue(String key, Object value, int expiration) {
        Element element = new Element(key, value);
        element.setTimeToLive(expiration);
        getCache().put(element);
    }

    private Cache getCache() {
        if (cache == null || cache.getStatus() != Status.STATUS_ALIVE) {
            CacheManager cacheManager = CacheManager.create();
            if (!cacheManager.cacheExists("psec")) {
                cacheManager.addCache("psec");
            }
            this.cache = cacheManager.getCache("psec"); 
        }
        return cache;
    }
    
    private Session getSession() {
        return Context.current().session();
    }


    private Response getResponse() {
        return Context.current().response();
    }


    private Request getRequest() {
        return Context.current().request();
    }

    
}
