package de.ppi.psec.services;

import java.security.Permission;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import models.psec.Role;

import play.Logger;
import play.Logger.ALogger;
import play.Play;
import de.ppi.psec.permissions.PsecBasicPermission;
import de.ppi.psec.permissions.PsecPermissionFactory;



public class AuthorizationServiceImpl implements AuthorizationService {

    private static final ALogger log = Logger.of(AuthorizationServiceImpl.class);

    private static AuthorizationService service;
    
    private final AuthenticationService authenticationService;
    
    private final RolePersistanceService rolePersistanceService;

    private boolean isEnabled = false;

    private final Map<RolePermissionKey, Boolean> rolePermissionMap =
            new HashMap<RolePermissionKey, Boolean>();

    public AuthorizationServiceImpl(AuthenticationService authenticationService) {        
        this(authenticationService, new RolePersistanceServiceEbean(),
                Play.application().isProd() ||
                Play.application().configuration().getBoolean("psecEnabled").booleanValue());
    }

    public AuthorizationServiceImpl(AuthenticationService authenticationService, 
            RolePersistanceService roleService, boolean isEnabled) {
        this.isEnabled = isEnabled;
        if (!isEnabled) {
            log.warn("psec is disabled");
        }
        this.authenticationService = authenticationService;
        this.rolePersistanceService = roleService;        
    }



    /**
     * {@inheritDoc}
     */
    public void clearPermissionCache(String reason) {
        if (!rolePermissionMap.isEmpty()) {
            if (log.isDebugEnabled()) {
                log.debug("Clear the cache (" + rolePermissionMap.size() +
                        "), because " + reason);
            }
            rolePermissionMap.clear();
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean hasRole(String roleName) {
        return authenticationService.getRoles().contains(roleName);
    }

    /**
     * {@inheritDoc}
     */
    public boolean hasPermission(Permission requiredPermission) {
        if (!isEnabled) {
            return true;
        }
        if (log.isTraceEnabled()) {
            log.trace("Required Permission "+ requiredPermission);
        }
        final List<Permission> userpermissions = authenticationService.getPermissions();
        for (Permission permission : userpermissions) {
            if (log.isTraceEnabled()) {
                log.trace("Check User-Permission "+ permission);
            }
            if (permission.implies(requiredPermission)) {
                if (log.isTraceEnabled()) {
                    log.trace("User-Permission implies Required-Permission");
                }
                return true;
            }

        }
        List<String> rolenames = authenticationService.getRoles();
        for (String rolename : rolenames) {
            if (log.isTraceEnabled()) {
                log.trace("Check permission for Role ${rolename}");
            }
            if (checkRolePermission(rolename, requiredPermission).booleanValue()) {
                if (log.isTraceEnabled()) {
                    log.trace("Role-Permission implies Required-Permission");
                }
                return true;
            }
        }
        return false;
    }

    /**
     * {@inheritDoc}
     */
    public boolean hasPermission(String targetName, String actionName) {
        return this.hasPermission(PsecPermissionFactory.
                getPermission(PsecBasicPermission.class.getName(),
                targetName != null ? targetName : "*",
                actionName != null ? actionName : "index"));
    }

    private Boolean checkRolePermission(String rolename, Permission requiredPermission) {
        if (rolename == null) {
            return Boolean.FALSE;
        }
        if (requiredPermission == null) {
            return Boolean.TRUE;
        }
        final RolePermissionKey key = new RolePermissionKey(rolename,
                requiredPermission);
        Boolean result = rolePermissionMap.get(key);
        if (result == null) {
            Role role = rolePersistanceService.findRoleByName(rolename);
            if (role == null) {
                log.warn("The role '" + rolename + " wasn't found.");
                result = Boolean.FALSE;
            } else {
                final List<Permission> permissions = role.getPermissions();
                result = Boolean.FALSE;
                for (Permission rolePermission : permissions) {
                    if (log.isTraceEnabled()) {
                        log.trace("Check Role-Permission "+ rolePermission);
                    }
                    if (rolePermission.implies(requiredPermission)) {
                        rolePermissionMap.put(key, Boolean.TRUE);
                        if (log.isTraceEnabled()) {
                            log.trace("Role-Permission implies Required-Permission");
                        }
                        result = Boolean.TRUE;
                        break;
                    }
                }
            }
        }

        if (log.isTraceEnabled()) {
            log.trace(rolename + " result " + result);
        }
        return result;
    }
    

    private static class RolePermissionKey {
        private final String roleName;
        private final Permission permission;

        public RolePermissionKey(String roleName, Permission permission) {
            super();
            this.roleName = roleName;
            this.permission = permission;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result
                    + ((permission == null) ? 0 : permission.hashCode());
            result = prime * result
                    + ((roleName == null) ? 0 : roleName.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            RolePermissionKey other = (RolePermissionKey) obj;
            if (permission == null) {
                if (other.permission != null) {
                    return false;
                }
            } else if (!permission.equals(other.permission)) {
                return false;
            }
            if (roleName == null) {
                if (other.roleName != null) {
                    return false;
                }
            } else if (!roleName.equals(other.roleName)) {
                return false;
            }
            return true;
        }

    }


    public static AuthorizationService getService() {
        if (service == null) {
            service = new AuthorizationServiceImpl(
                    AuthenticationServiceImpl.getService());
        }
        return service;
    }




    public synchronized static void setService(AuthorizationService service) {
        AuthorizationServiceImpl.service = service;
    }
}


