/**
 * 
 */
package de.ppi.psec.services;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import play.Logger;
import play.Play;
import play.Logger.ALogger;
import play.mvc.Result;


/**
 * Small Helper Class to handle public actions by configuration.
 * @author niels
 *
 */
public class PublicActionsByConfiguration {
    private final static ALogger log = Logger.of(PublicActionsByConfiguration.class);
    
    
    private static long lastConfigDetectionTime = 0;

    private static Set<Method> publicByConfigurationsMethods = new HashSet <Method>();
    
    private static Enumeration<URL> publicActionsUrls; 
    
    public static boolean isPublic(Method m) {
        try {
            initPublicMethodsByConfiguration();
        } catch (IOException e) {
            throw new IllegalStateException(
                    "Public Methods should have been initialized", e);
        }
        return publicByConfigurationsMethods.contains(m);
    }
    
    public static void initPublicMethodsByConfiguration() throws IOException {
        boolean scan = !Play.isProd() || publicActionsUrls == null;
        publicActionsUrls = Play.application().classloader().
                getResources("psec.publicActions");
        List<URL> changedUrls = new ArrayList <URL>();
        while (scan && publicActionsUrls.hasMoreElements()) {
            final URL url = publicActionsUrls.nextElement();
            final long modificationTime = new File(url.getFile()).lastModified();
            if (modificationTime > lastConfigDetectionTime) {
                changedUrls.add(url);
            }
        }
        if (!changedUrls.isEmpty()) {
            loadPublicMethodsByConfiguration(changedUrls);
        }
    }
    
    @SuppressWarnings("rawtypes")
    private static void loadPublicMethodsByConfiguration(List<URL> urls) {
        log.info("Loading public actions.");
        lastConfigDetectionTime = System.currentTimeMillis();
        publicByConfigurationsMethods.clear();
        for (URL url : urls) {
            log.info("Found public action descriptor, " + url);
            try {
                BufferedReader reader = new BufferedReader(new InputStreamReader(url.openStream(), "utf-8"));
                String line = null;
                while ((line = reader.readLine()) != null) {
                    line = line.trim();
                    if (line.length() == 0 || line.startsWith("#")) {
                        continue;
                    }
                    int methodDelimiter = line.lastIndexOf('.');
                    if (methodDelimiter <= 1) {
                        Logger.warn(">" + line + "< has an invalid structure.");
                        continue;
                    }
                    final String controllerClassName = line.substring(0, methodDelimiter);
                    try {
                        Class controller = Play.application().classloader().loadClass("controllers." + controllerClassName);
                        final String methodName = line.substring(methodDelimiter+1);
                        Class subController = controller; 
                        if (methodName.equals("*")) {
                            while (!subController.getName().equals("java.lang.Object")) {
                                for (Method m : subController.getDeclaredMethods()) {
                                    addPublicMethodByConfiguration(m);
                                }
                                subController = subController.getSuperclass();
                            }                        
                        } else {
                            final Method[] methods = subController.getMethods();
                            for (Method m : methods) {
                                if (m.getName().equals(methodName)) {
                                    addPublicMethodByConfiguration(m);
                                    break;
                                }
                            }
                        } 
                    } catch (ClassNotFoundException e) {
                        if (!"TestRunner".equals(controllerClassName)) {
                            log.error("Class 'controllers." + controllerClassName + 
                                    "' can't be loaded.");
                        }
                    }
                }
            } catch (Exception ex) {
                log.error("Cannot load %s " + url, ex);
            }
        }
    }
    
    private static boolean isActionMethod(Method m) {
        final int mod = m.getModifiers();
        return Modifier.isPublic(mod) &&
                Modifier.isStatic(mod) &&
                Result.class.isAssignableFrom(m.getReturnType());
    }
    
    private static void addPublicMethodByConfiguration(Method method) {
        if (isActionMethod(method)) {
            log.info("Method " + method + " is public by configuration");
            publicByConfigurationsMethods.add(method);
        }
    }
    


}
