package de.ppi.util;

import java.io.UnsupportedEncodingException;
import java.security.SecureRandom;
import java.util.Arrays;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.codec.digest.DigestUtils;

/**
 * Encapsulates several password-related utility functions Some functions came
 * originally from http://www.securitydocs.com/library/3439
 */
public class PasswordTools {

    private static final String ALLOWED_CHARACTERS =
            "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!@,;.-=+";

    private static final String LOGIN_CHARACTERS =
            "0123456789abcdefghijklmnopqrstuvwxyz.";

    private static final Hex CODEC = new Hex();

    public static String generateRandomPassword() {
        return generateRandomPassword(10, ALLOWED_CHARACTERS);
    }

    /**
     * Generates a random password
     *
     * @param size
     *            Password size
     * @returns Random password
     */
    public static String generateRandomPassword(int size) {
        return generateRandomPassword(size, ALLOWED_CHARACTERS);
    }

    /**
     * Generates a random password out of a alowed characters
     *
     * @param size
     *            Password size
     * @param allowed
     *            String comprising the allowed characters
     * @returns Random password
     */
    public static String generateRandomPassword(int size, String allowed) {
        SecureRandom random = new SecureRandom();

        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < size; i++) {
            int rand = random.nextInt(allowed.length());
            sb.append(allowed.substring(rand, rand + 1));
        }
        return sb.toString();
    }

    public static String generateRandomLogin() {
        return generateRandomLogin(10);
    }

    /**
     * Returns a random string made up only of numbers, letters and the period
     * character
     */
    public static String generateRandomLogin(int size) {
        return generateRandomPassword(size, LOGIN_CHARACTERS);
    }

    public static byte[] generateSalt() {
        return generateSalt(40);
    }

    /**
     * Generates a random salt of a certain size
     *
     * @param size
     *            How many bytes should be in the salt
     */
    public static byte[] generateSalt(int size) {
        SecureRandom random = new SecureRandom();
        final byte[] list = new byte[size];
        for (int i = 0; i < list.length; i++) {
            list[i] = (byte) random.nextInt();
        }
        return list;
    }

    /**
     * Combine two byte arrays
     *
     * @param l
     *            first byte array
     * @param r
     *            second byte array
     * @return byte[] combined byte array
     */
    private static byte[] concatenate(byte[] left, byte[] right) {
        byte[] b = new byte[left.length + right.length];
        System.arraycopy(left, 0, b, 0, left.length);
        System.arraycopy(right, 0, b, left.length, right.length);
        return b;
    }

    /**
     * split a byte array in two
     *
     * @param src
     *            byte array to be split
     * @param n
     *            element at which to split the byte array
     * @return byte[][] two byte arrays that have been split
     */
    private static byte[][] split(byte[] src, int n) {
        byte[] l, r;
        if (src == null || src.length <= n) {
            l = src;
            r = new byte[0];
        } else {
            l = new byte[n];
            r = new byte[src.length - n];
            System.arraycopy(src, 0, l, 0, n);
            System.arraycopy(src, n, r, 0, r.length);
        }
        byte[][] lr = { l, r };
        return lr;
    }

    /**
     * SHA512 password and a random salt with 2^5 rounds.
     *
     * @param password
     *            String to hash.
     * @return byte array concatenating the 64-byte hash and the salt.
     *
     */
    public static byte[] saltPassword(String password) {
        return saltPassword(password, HashMethod.SHA512, (byte) 5);
    }

    /**
     * Hash password and a random salt. The hash-method will be called
     * <code>2^roundExponent</code> times.
     *
     * @param password
     *            password which should be hashed and salted.
     * @param hashMethod
     *            the used method to hash the values.
     * @param roundExponent
     *            roundExponent defines how often the hash-value will be
     *            rehashed.
     * @return the hashed password.
     *
     */
    public static byte[] saltPassword(String password, HashMethod hashMethod,
            byte roundExponent) {
        final byte[] method = { hashMethod.getId(), roundExponent };
        final byte[] salt =
                generateSalt(hashMethod.getHasher().getHashLength());
        final byte[] pwdBytes = password.getBytes();
        final Hasher hasher = hashMethod.getHasher();
        final byte[] hash =
                hasher.hash(concatenate(pwdBytes, salt), roundExponent);
        return concatenate(method, concatenate(hash, salt));
    }

    /**
     * Returns a salted password base64 encoded.
     *
     * @see PasswordTools#saltPassword(String, HashMethod, byte)
     */
    public static String saltPasswordBase64(String password,
            HashMethod hashMethod, byte roundExponent) {
        final String encoded =
                new String(Base64.encodeBase64(saltPassword(password,
                        hashMethod, roundExponent)));
        return encoded;
    }

    /**
     * Returns a salted password hex-encoded.
     *
     * @see PasswordTools#saltPassword(String, HashMethod, byte).
     */
    public static String saltPasswordHex(String password,
            HashMethod hashMethod, byte roundExponent) {
        final String encoded =
                new String(CODEC.encode(saltPassword(password, hashMethod,
                        roundExponent)));
        return encoded;
    }

    /**
     * SHA512 password and a random salt with 2^5 rounds.
     *
     * @param password
     *            String to hash
     * @return String Base64-encoded byte array concatenating the 64-byte hash
     *         and the salt
     */
    public static String saltPasswordBase64(String password) {
        final String encoded =
                new String(Base64.encodeBase64(saltPassword(password)));
        return encoded;
    }

    /**
     * SHA512 password and a random salt with 2^5 rounds.
     *
     * @param password
     *            String to hash
     * @return String Hex-encoded byte array concatenating the 64-byte hash and
     *         the salt
     */
    public static String saltPasswordHex(String password) {
        final String encoded = new String(CODEC.encode(saltPassword(password)));
        return encoded;
    }

    /**
     * Verifies a password against a hex-encoded digest.
     *
     * @param password
     *            Value to verify
     * @param digestBase64
     *            byte array concatenating the hash and the salt and in the
     *            first 2 bytes the hash-method in based64.
     * @return <code>true</code> if successful
     *
     * @throws DecoderException
     * @see PasswordTools#checkDigest(String,byte[])
     */
    public static boolean checkDigestHex(String password, String digestHex)
            throws DecoderException {
        final byte[] digest = CODEC.decode(digestHex.getBytes());
        return checkDigest(password, digest);
    }

    /**
     * Verifies a password against a base64-encoded digest
     *
     * @param password
     *            Value to verify
     * @param digestBase64
     *            byte array concatenating the hash and the salt and in the
     *            first 2 bytes the hash-method in based64.
     * @return <code>true</code> if successful
     * @throws UnsupportedEncodingException
     * @see PasswordTools#checkDigest(String,byte[])
     */
    public static boolean checkDigestBase64(String password, String digestBase64)
            throws UnsupportedEncodingException {
        final byte[] digest =
                Base64.decodeBase64(digestBase64.getBytes("utf-8"));
        return checkDigest(password, digest);
    }

    /**
     * Verifies a password against a digest
     *
     * @param password
     *            Value to verify
     * @param digestWithMethod
     *            byte array concatenating the hash and the salt and in the
     *            first 2 bytes the hash-method.
     * @return <code>true</code> if successful
     *
     */
    public static boolean checkDigest(String password, byte[] digestWithMethod) {

        if (digestWithMethod == null) {
            return password == null;
        }
        if (digestWithMethod.length < 2) {
            throw new IllegalArgumentException(
                    "The digest must be at least 2 bytes");
        }
        final Hasher hasher =
                HashMethod.getInstance(digestWithMethod[0]).getHasher();
        final int roundExponent = digestWithMethod[1];
        final byte[] digest = new byte[digestWithMethod.length - 2];
        System.arraycopy(digestWithMethod, 2, digest, 0, digest.length);

        final int byteLength = hasher.getHashLength();
        // First section will contain the original hash, the next the salt
        final byte[][] hs = split(digest, byteLength);
        final byte[] hash = hs[0];
        final byte[] salt = hs[1];

        // Update digest object with byte array of clear text string and salt
        final byte[] concat = concatenate(password.getBytes(), salt);

        // Complete hash computation, this is now binary data
        final byte[] pwhash;
        pwhash = hasher.hash(concat, roundExponent);

        final boolean valid = Arrays.equals(hash, pwhash);

        return valid;
    }

    /**
     * List of possible hash-methods.
     *
     */
    public static enum HashMethod {
        SHA(1, new SHA()), SHA256(2, new SHA256()), SHA512(3, new SHA512());

        private final byte id;
        private final Hasher hasher;

        private HashMethod(int id, Hasher hasher) {
            this.id = (byte) id;
            this.hasher = hasher;
        }

        /**
         * The id of the method.
         *
         * @return id of the method.
         */
        public byte getId() {
            return id;
        }

        /**
         * The object which do the real hash-work.
         *
         * @return object which do the real hash-work.
         */
        public Hasher getHasher() {
            return hasher;
        }

        /**
         * Deliver an instance for the given id.
         *
         * @param id
         *            the id of the wanted method.
         * @return an instance for the given id.
         */
        public static HashMethod getInstance(byte id) {
            for (HashMethod method : HashMethod.values()) {
                if (method.getId() == id) {
                    return method;
                }
            }
            throw new IllegalArgumentException("Id " + id + " is unknown.");
        }

    }

    /**
     * Class which hash a value many times.
     *
     *
     */
    public static abstract class Hasher {

        /**
         * Hash the data with 2^roundExponent-rounds
         *
         * @param data
         *            the data.
         * @param roundExponent
         *            defines how often the hash-value will be rehashed.
         * @return the hashvalue.
         */
        public byte[] hash(byte[] data, int roundExponent) {
            final long rounds = 2 ^ roundExponent;
            byte[] hashValue = hash(data);
            for (int i = 1; i < rounds; i++) {
                hashValue = hash(hashValue);
            }
            return hashValue;
        }

        /**
         * Hash the data once.
         *
         * @param data
         *            the data.
         * @return the hashvalue.
         */
        protected abstract byte[] hash(byte[] data);

        /**
         * The length of the hashvalue.
         *
         * @return length of the hashvalue.
         */
        public abstract int getHashLength();

    }

    public static class SHA512 extends Hasher {

        @Override
        protected byte[] hash(byte[] data) {
            return DigestUtils.sha512(data);
        }

        @Override
        public int getHashLength() {
            return 64;
        }
    }

    public static class SHA256 extends Hasher {

        @Override
        protected byte[] hash(byte[] data) {
            return DigestUtils.sha256(data);
        }

        @Override
        public int getHashLength() {
            return 32;
        }
    }

    public static class SHA extends Hasher {

        @Override
        protected byte[] hash(byte[] data) {
            return DigestUtils.sha(data);
        }

        @Override
        public int getHashLength() {
            return 20;
        }

    }

}
