package de.ppi.psec.services;

import static org.fest.assertions.Assertions.assertThat;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.security.Permission;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import models.psec.PsecPermission;
import models.psec.PsecRole;
import models.psec.PsecUser;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import play.mvc.Http;
import play.mvc.Http.Context;
import play.mvc.Http.Cookie;
import play.mvc.Http.Request;
import play.mvc.Http.Response;
import play.mvc.Http.Session;
import de.ppi.psec.permissions.PsecBasicPermission;

@SuppressWarnings("boxing")
public class AuthenticationServiceTest {

    private AuthenticationService service;

    private static final String SESSION_KEY_USERID = "psecUid";
    private static final String SESSION_KEY_REMEMBERED = "psecRem";
    private static final String COOKIE_NAME = "psec_rememberMe";
    
    Request requestMock; 
    Http.Cookies cookiesMock;
    UserPersistanceService userPersistanceServiceMock = mock(UserPersistanceService.class);
    
    @Before
    public void setUp() {
        service = new AuthenticationServiceImpl(10, "100d", "abcdefghijklmnopqrst", 
                userPersistanceServiceMock);
        initSessionRequestAndResponse();
    }
    
    private void initSessionRequestAndResponse() {
        requestMock = mock(Request.class); 
        cookiesMock = mock(Http.Cookies.class);
        when(requestMock.cookies()).thenReturn(cookiesMock);
        Context.current.set(new Context(requestMock, new HashMap <String, String>(), 
                new HashMap <String, String>()));
        
    }

    @BeforeClass
    public static void setUpBeforeClass() throws Exception {
    }


    @Test
    public void testIsAuthenticate() {
        assertFalse(service.isAuthenticate(true));
        getSession().put(SESSION_KEY_USERID, "1");
        assertTrue(service.isAuthenticate(false));
        getSession().put(SESSION_KEY_REMEMBERED, "1");
        assertTrue(service.isAuthenticate(false));
        getSession().put(SESSION_KEY_REMEMBERED, "Y");
        assertFalse(service.isAuthenticate(false));
    }

   
    
    @Test
    public void testCanRemembered() {
        boolean[] booleans = {true, false};
        final PsecUser psecUser = new PsecUser();
        psecUser.setId(Long.valueOf(1));
        psecUser.setLoginname("hugo");
        psecUser.setPassword("secret");
        psecUser.setLoginFailedCount(0);
        psecUser.setEnabled(true);
        when(userPersistanceServiceMock.findUserById(psecUser.getId())).
                thenReturn(psecUser);
        when(userPersistanceServiceMock.findUserByLoginname(psecUser.getLoginname())).
                thenReturn(psecUser);
        for (boolean remember : booleans) {
            for (boolean js : booleans) {
                assertTrue(service.login("hugo", "secret", remember, js));
                copyCookiesFromResponseToRequest();
                assertEquals(remember, service.canRemembered());
                initSessionRequestAndResponse();
            }
        }
        assertTrue(service.login("hugo", "secret", true, true));
        copyCookiesFromResponseToRequest();
        assertTrue(service.canRemembered());
        service.logout();
        assertNotNull(getDiscardedCookie(COOKIE_NAME));
        when(cookiesMock.get(COOKIE_NAME)).thenReturn(null);
        assertFalse(service.canRemembered());
        
    }

    private void copyCookiesFromResponseToRequest() {
        final Iterable <Cookie> cookies = getResponse().cookies();
        for (Cookie cookie : cookies) {
            when(cookiesMock.get(cookie.name())).thenReturn(cookie);
        }
    }
    
    @Test
    public void testLogin() {
        boolean[] booleans = {false, true};
        final PsecUser psecUser = new PsecUser();
        psecUser.setId(Long.valueOf(1));
        psecUser.setLoginname("hugo");
        psecUser.setPassword("secret");
        psecUser.setLoginFailedCount(0);
        psecUser.setEnabled(true);
        when(userPersistanceServiceMock.findUserById(psecUser.getId())).
                thenReturn(psecUser);
        when(userPersistanceServiceMock.findUserByLoginname(psecUser.getLoginname())).
                thenReturn(psecUser);

        for (boolean remember : booleans) {
            for (boolean js : booleans) {
            	assertNull(getCookie(COOKIE_NAME));
                assertTrue(service.login("hugo", "secret", remember, js));                
                assertEquals(remember, getCookie(COOKIE_NAME) != null);
                if  (remember) {                
                    final Cookie rememberMeCookie = getCookie(COOKIE_NAME);
                    assertEquals(100*24*60*60, rememberMeCookie.maxAge());
                }
                assertEquals("1",Context.current().session().get(SESSION_KEY_USERID));
                assertFalse(Context.current().session().containsKey(SESSION_KEY_REMEMBERED));
                initSessionRequestAndResponse();
                //login failed
                //passwort wrong
                assertFalse(service.login("hugo", "secrets", remember, js));
                //loginname wrong
                assertFalse(service.login("hugos", "secret", remember, js));
                //user disabled.
                psecUser.setEnabled(false);
                assertFalse(service.login("hugos", "secret", remember, js));
                psecUser.setEnabled(true);
                
                //test failCount
                psecUser.setLoginFailedCount(10);
                assertTrue(service.login("hugo", "secret", remember, js));
                assertEquals(0, psecUser.getLoginFailedCount());
                psecUser.setLoginFailedCount(10);
                assertFalse(service.login("hugo", "secrets", remember, js));
                assertEquals(11, psecUser.getLoginFailedCount());
                assertFalse(service.login("hugo", "secret", remember, js));
                psecUser.setLoginFailedCount(0);
                initSessionRequestAndResponse();
            }
        }
    }

    private Cookie getCookie(String name) {
    	final Iterable<Cookie> cookies = getResponse().cookies();
    	for (Cookie cookie : cookies) {
			if (name.equals(cookie.name())) {
				return cookie;
			}
    	}
    	return null;
    }
    

    private String getDiscardedCookie(String name) {
        final Iterable<String> cookieNames = getResponse().discardedCookies();
        for (String cookieName: cookieNames) {
            if (name.equals(cookieName)) {
                return cookieName;
            }
        }
        return null;
    }
    
    @Test
    public void testLogout() {        
        getSession().put(SESSION_KEY_USERID, "1");
        getResponse().setCookie(COOKIE_NAME, "A value");
        assertNotNull(getSession().get(SESSION_KEY_USERID));
        Cookie cookie = getCookie(COOKIE_NAME);
        assertNotNull(cookie);        
        assertThat(cookie.maxAge()).isNotEqualTo(0);
        assertEquals("A value", cookie.value());
        when(cookiesMock.get(COOKIE_NAME)).thenReturn(cookie);       
        service.logout();
        assertNull(getSession().get(SESSION_KEY_USERID));
        assertNotNull(getDiscardedCookie(COOKIE_NAME));
    }

    
    @Test
    public void testSessionValues() {
        final PsecUser psecUser = new PsecUser();
        psecUser.setId(Long.valueOf(1));
        psecUser.setLoginname("hugo");
        psecUser.setPassword("secret");
        psecUser.setFirstname("Hugo");
        psecUser.setLastname("Boss");
        psecUser.setLoginFailedCount(0);
        psecUser.setEnabled(true);

        final List <String> roleNames = new ArrayList <String>();
        final List <String> permNames = new ArrayList <String>();
        final List <PsecPermission> permissions = new ArrayList <PsecPermission>();
        for (int i = 0; i < 10; i++) {
            final  String roleName = "role_" + i;
            final String permName = "perm_" + i;
            roleNames.add(roleName);
            permNames.add(permName);
            final PsecRole role = new PsecRole();    
            role.setName(roleName);
            final PsecPermission perm = createPermission(permName);    
            final List <PsecPermission> rolePermissions = new ArrayList <PsecPermission>();
            final PsecPermission permRole = createPermission(roleName);    
            rolePermissions.add(permRole);
            role.setPsecPermissions(rolePermissions); 
            permissions.add(perm);
            psecUser.addToRoles(role);
        }
        psecUser.setPsecPermissions(permissions);
        
        when(userPersistanceServiceMock.findUserById(psecUser.getId())).
                thenReturn(psecUser);
        when(userPersistanceServiceMock.findUserByLoginname(psecUser.getLoginname())).
                thenReturn(psecUser);

        assertTrue(service.login("hugo", "secret", true, true));
        
        assertEquals(psecUser.getFirstname(), service.getFirstname());
        assertEquals(psecUser.getLastname(), service.getLastname());
        assertEquals(psecUser.getLoginname(), service.getLoginname());
        assertEquals(psecUser.getId().toString(), service.getUserid());
        final List<String> serviceRoleList = service.getRoles();
        assertEquals(roleNames.size(), serviceRoleList.size());
        for (int i = 0; i < roleNames.size(); i++) {
            assertEquals(roleNames.get(i), serviceRoleList.get(i));
        }
        final List<Permission> servicePermList = service.getPermissions();
        assertEquals(permissions.size(), servicePermList.size());
        for (int i = 0; i < permissions.size(); i++) {
            assertEquals(permissions.get(i).convertToPermission(), servicePermList.get(i));
        }
    }
    
    private PsecPermission createPermission(String name) {
        PsecPermission result = new PsecPermission();
        result.setName(name);
        result.setType(PsecBasicPermission.class.getName());
        result.setActions(name);
        result.setTarget(name);
        return result;
    }
    
    private Session getSession() {
        return Context.current().session();
    }
       
    private Response getResponse() {
        return Context.current().response();
    }
        
}
