package de.ppi.util;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;

import junit.framework.TestCase;

import org.apache.commons.codec.DecoderException;

import de.ppi.util.PasswordTools.HashMethod;

public class PasswordToolsTest extends TestCase {

    public void testSalt() {
        int size = 4;
        List<String> salts = new ArrayList<String>();
        for (int i = 0;  i < 100; i++) {
            String salt = PasswordTools.generateRandomPassword(size);
            assertEquals(-1, salts.indexOf(salt));
            salts.add(salt);
        }
        assertEquals(100, salts.size());
        for (String salt : salts) {
        	assertEquals(size, salt.length());
		}
        size = 20;
        List<byte[]> bsalts = new ArrayList<byte[]>();
        for (int i = 0;  i < 10; i++) {
            bsalts.add(PasswordTools.generateSalt(size));
        }
        assertEquals(10, bsalts.size());
        for (byte[] bs : bsalts) {
        	assertEquals(size, bs.length);
		}
    }

    public void testSaltedPasswordHex() throws DecoderException {
        String password = "superSecret";
        String salted = PasswordTools.saltPasswordHex(password);
        String again = PasswordTools.saltPasswordHex(password);

        assertFalse(salted.equals(again));

        assertTrue(PasswordTools.checkDigestHex(password, salted));
        assertTrue(PasswordTools.checkDigestHex(password, again));
        assertFalse(PasswordTools.checkDigestHex("SuperSecret", salted));

        List<String> hashes = new ArrayList<String>();
        for (int i = 0;  i < 100; i++) {
            // Verify we don't get duplicated hashes
            salted = PasswordTools.saltPasswordHex(password);
            assertEquals(-1, hashes.indexOf(salted));
            hashes.add(salted);
        }
        for (String hash : hashes) {
            assertTrue(PasswordTools.checkDigestHex(password, hash));
            assertFalse(PasswordTools.checkDigestHex("SuperSecret", hash));
        }
        for (HashMethod method : PasswordTools.HashMethod.values()) {
            for (byte i = 0; i < 5; i++) {
                internalTestSaltedPasswordHex(method, i);
            }
        }

    }

    public void testSaltedPasswordBase64() throws UnsupportedEncodingException {
        String password = "superSecret";
        String salted = PasswordTools.saltPasswordBase64(password);
        String again = PasswordTools.saltPasswordBase64(password);

        assertFalse(salted.equals(again));

        assertTrue(PasswordTools.checkDigestBase64(password, salted));
        assertTrue(PasswordTools.checkDigestBase64(password, again));
        assertFalse(PasswordTools.checkDigestBase64("SuperSecret", salted));

        List<String> hashes = new ArrayList<String>();
        for (int i = 0;  i < 100; i++) {
            // Verify we don't get duplicated hashes
            salted = PasswordTools.saltPasswordBase64(password);
            assertEquals(-1, hashes.indexOf(salted));
            hashes.add(salted);
        }

        for (String hash : hashes) {
            assertTrue(PasswordTools.checkDigestBase64(password, hash));
            assertFalse(PasswordTools.checkDigestBase64("SuperSecret", hash));
        }
        for (HashMethod method : PasswordTools.HashMethod.values()) {
            for (byte i = 0; i < 5; i++) {
                internalTestSaltedPasswordBase64(method, i);
            }
        }
    }




    private void internalTestSaltedPasswordHex(HashMethod method, byte roundUpExponent) throws DecoderException {
        int passwordLength = 2^roundUpExponent;
        final String password = "a" + PasswordTools.generateRandomLogin(passwordLength);
        passwordLength++;
        assertEquals(passwordLength, password.length());
        String salted = PasswordTools.saltPasswordHex(password, method, roundUpExponent);
        final String again = PasswordTools.saltPasswordHex(password, method, roundUpExponent);
        assertFalse(salted.equals(again));
        assertTrue(PasswordTools.checkDigestHex(password, salted));
        assertTrue(PasswordTools.checkDigestHex(password, again));
        final String wrongPassword = "A" + password.substring(1);
        assertFalse(password + " !=" + wrongPassword ,
                PasswordTools.checkDigestHex(wrongPassword, salted));
        List<String> hashes = new ArrayList<String>();
        for (int i = 0;  i < 100; i++) {
            // Verify we don't get duplicated hashes
            salted = PasswordTools.saltPasswordHex(password, method, roundUpExponent);
            assertEquals(-1, hashes.indexOf(salted));
            hashes.add(salted);
        }
        for (String hash : hashes) {
            assertTrue(PasswordTools.checkDigestHex(password, hash));
            assertFalse(PasswordTools.checkDigestHex("SuperSecret", hash));
        }
    }

    private void internalTestSaltedPasswordBase64(HashMethod method, byte roundUpExponent) throws UnsupportedEncodingException {
        int passwordLength = 2^roundUpExponent;
        final String password = "a" + PasswordTools.generateRandomPassword(passwordLength);
        passwordLength++;
        assertEquals(passwordLength, password.length());
        String salted = PasswordTools.saltPasswordBase64(password, method, roundUpExponent);
        String again = PasswordTools.saltPasswordBase64(password, method, roundUpExponent);

        assertFalse(salted.equals(again));

        assertTrue(PasswordTools.checkDigestBase64(password, salted));
        assertTrue(PasswordTools.checkDigestBase64(password, again));
        final String wrongPassword = "A" + password.substring(1);
        assertFalse(password + " !=" + wrongPassword ,
                PasswordTools.checkDigestBase64(wrongPassword, salted));

        List<String> hashes = new ArrayList<String>();
        for (int i = 0;  i < 100; i++) {
            // Verify we don't get duplicated hashes
            salted = PasswordTools.saltPasswordBase64(password, method, roundUpExponent);
            assertEquals(-1, hashes.indexOf(salted));
            hashes.add(salted);
        }

        for (String hash : hashes) {
            assertTrue(PasswordTools.checkDigestBase64(password, hash));
            assertFalse(PasswordTools.checkDigestBase64("SuperSecret", hash));
        }

    }

}
